@@ -1,11 +1,11 @@
 from option_merge.merge import ConverterProperty, KeyValuePairsConverter, AttributesConverter
-from option_merge.merge import MergedOptions, BadPrefix
+from option_merge.merge import MergedOptions
 from option_merge.converter import Converter
-from option_merge.helper import NotFound
+from option_merge.not_found import NotFound
 
 # Explicitly make these part of this module
 NotFound = NotFound
-BadPrefix = BadPrefix
+Converter = Converter
 MergedOptions = MergedOptions
 ConverterProperty = ConverterProperty
 AttributesConverter = AttributesConverter
