@@ -11,18 +11,17 @@ With the ability to delete from the dictionary and the ability to convert values
 on access.
 """
 
-from option_merge.converter import Converters
+from option_merge.versioning import versioned_iterable, versioned_value, VersionedDict
+from option_merge.converter import Converters, Converter
+from option_merge.not_found import NotFound
 from option_merge.joiner import dot_joiner
-from option_merge.storage import Storage
-from option_merge import helper as hp
 from option_merge.path import Path
 
-from delfick_error import DelfickError
 from collections import Mapping
+import logging
 import six
 
-class BadPrefix(DelfickError): pass
-"""Used to say a prefix value is not a dictionary"""
+log = logging.getLogger("option_merge.merge")
 
 class KeyValuePairsConverter(object):
     """Converts a list of key,value pairs to a dictionary"""
@@ -32,6 +31,7 @@ class KeyValuePairsConverter(object):
 
     def convert(self):
         """Return us a MergedOptions from our pairs"""
+        from option_merge import helper as hp
         return MergedOptions().using(*[hp.make_dict(key[0], key[1:], value) for key, value in self.pairs], source=self.source)
 
 class AttributesConverter(object):
@@ -126,15 +126,33 @@ class MergedOptions(dict, Mapping):
             self.converters = Converters()
         if isinstance(self.prefix_list, six.string_types):
             self.prefix_list = [self.prefix_list]
-        self.prefix_string = dot_joiner(self.prefix_list)
+        self.prefix_string = dot_joiner(self.prefix_list, list)
 
         self.storage = storage
+        from option_merge.storage import Storage
         if self.storage is None:
             self.storage = Storage()
 
     @classmethod
     def using(cls, *options, **kwargs):
-        """Convenience for calling update multiple times"""
+        """
+        Convenience for calling update multiple times
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1}, {"b": 2})
+
+        is equivalent to:
+
+        .. code-block:: python
+
+            m = MergedOptions()
+            m.update({"a": 1})
+            m.update({"b": 2})
+
+        Any kwargs given to ``using`` is passed into ``update`` for
+        each provided dictionary.
+        """
         prefix = kwargs.get('prefix')
         storage = kwargs.get('storage')
         converters = kwargs.get('converters')
@@ -149,17 +167,34 @@ class MergedOptions(dict, Mapping):
             merged.update(opts, **kwargs)
         return merged
 
+    @property
+    def version(self):
+        return self.storage.version
+
     def update(self, options, source=None, **kwargs):
-        """Add new options"""
+        """
+        Add new options to the storage under this prefix.
+
+        The later options are added, the more influence they have.
+        """
         if options is None: return
         self.storage.add(Path(self.prefix_list), options, source=source)
 
+    @versioned_value
     def __getitem__(self, path, ignore_converters=False):
         """
         Access some path
 
         Return the first value it comes across
+
         Raise KeyError if nothing has the specified key
+
+        for example:
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": {"b": 2}})
+            assert m['a.b'] == 2
         """
         path = self.converted_path(path, ignore_converters=ignore_converters or self.ignore_converters or getattr(path, "ignore_converters", False))
         for val, return_as_is in self.values_for(path, ignore_converters=path.ignore_converters):
@@ -168,14 +203,24 @@ class MergedOptions(dict, Mapping):
 
             if any(isinstance(val, unprefixed) for unprefixed in self.dont_prefix):
                 return val
-            elif isinstance(val, dict):
+            elif type(val) in (dict, VersionedDict, MergedOptions) or isinstance(val, dict):
                 return self.prefixed(path, already_prefixed=True)
             else:
                 return val
+
         raise KeyError(path)
 
     def __contains__(self, path):
-        """Ask storage if it has a path"""
+        """
+        Ask storage if it has a path
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1})
+
+            assert "a" in m
+            assert "b" not in m
+        """
         try:
             self.storage.get(self.converted_path(path))
             return True
@@ -183,25 +228,78 @@ class MergedOptions(dict, Mapping):
             return False
 
     def get(self, path, default=None, ignore_converters=False):
-        """Get some path or return default value"""
+        """
+        Get some path or return default value
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1})
+
+            assert m.get("a") == 1
+            assert m.get("b") == None
+            assert m.get("b", 2) == 2
+
+        You may also specify ``ignore_converters`` and it won't take the
+        the converters into account.
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1})
+            m.add_converter(Converter(convert=add_one, convert_path=["a"]))
+            m.converters.activate()
+
+            assert m["a"] == 2
+            assert m.get("a", ignore_converters=True) == 1
+        """
         try:
             return self.__getitem__(path, ignore_converters=ignore_converters)
         except KeyError:
             return default
 
     def source_for(self, path, chain=None):
-        """Proxy self.storage.source_for"""
+        """
+        Proxy self.storage.source_for
+
+        Source is specifying in calls to ``__init__`` and ``update`` and this
+        will find the entries for the specified path an return the first source
+        it finds.
+        """
         path = Path.convert(path, self).ignoring_converters(True)
         return self.storage.source_for(path, chain)
 
     def __setitem__(self, path, value):
-        """Set a key in the storage"""
+        """
+        Set a key in the storage
+
+        This takes into account the prefix on this option as well as the provided
+        path.
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1})
+            assertEqual(m.as_dict(), {"a": 1})
+
+            a = m["a"]
+            a['b'] = 2
+
+            assertEqual(m.as_dict(), {"a": 1, "b": 2})
+        """
         if isinstance(path, six.string_types):
             path = [path]
         self.storage.add(self.converted_path(path), value)
 
     def __delitem__(self, path):
-        """Delete a key from the storage"""
+        """
+        Delete a key from the storage
+
+        .. code-block:: python
+
+            m = MergedOptions.using({"a": 1}, {"a": 2})
+            assert m['a'] == 2
+
+            del m['a']
+            assert m['a'] == 1
+        """
         self.storage.delete(self.converted_path(path))
 
     def __iter__(self):
@@ -216,9 +314,10 @@ class MergedOptions(dict, Mapping):
         """Equal to another merged options if has same storage and prefix"""
         return isinstance(other, self.__class__) and other.storage is self.storage and other.prefix_list == self.prefix_list
 
-    def values_for(self, path, converters=None, ignore_converters=False):
+    @versioned_iterable
+    def values_for(self, path, ignore_converters=False):
         """Get all known values for some path"""
-        path = self.converted_path(path, converters=converters, ignore_converters=ignore_converters or getattr(path, "ignore_converters", False))
+        path = self.converted_path(path, ignore_converters=ignore_converters or getattr(path, "ignore_converters", False))
         if not path.ignore_converters and not path.waiting():
             if path.converted():
                 yield path.converted_val(), True
@@ -232,7 +331,7 @@ class MergedOptions(dict, Mapping):
         for info in self.storage.get_info(path):
             try:
                 yield info.value_after(path), False
-            except hp.NotFound:
+            except NotFound:
                 pass
 
     def prefixed(self, path, ignore_converters=False, already_prefixed=False):
@@ -246,7 +345,7 @@ class MergedOptions(dict, Mapping):
             )
 
     def root(self):
-        """Return a MergedOptions prefixed to this path"""
+        """Return a MergedOptions looking at the root of the storage"""
         return self.__class__(
               ""
             , storage=self.storage
@@ -256,7 +355,16 @@ class MergedOptions(dict, Mapping):
             )
 
     def wrapped(self):
-        """Return a MergedOptions with this inside"""
+        """
+        Return a MergedOptions with this inside
+
+        Equivalent to:
+
+        .. code-block:: python
+
+            m = MergedOptions.using("a")
+            wrapped = MergedOptions.using(m, converters=m.converters, dont_prefix=m.dont_prefix)
+        """
         return self.__class__.using(self, converters=self.converters, dont_prefix=self.dont_prefix)
 
     def keys(self, ignore_converters=False):
@@ -269,8 +377,9 @@ class MergedOptions(dict, Mapping):
         for key in self.keys(ignore_converters=ignore_converters):
             yield key, self.__getitem__(key, ignore_converters=ignore_converters)
 
+    @versioned_iterable
     def values(self):
-        """Return the values"""
+        """Return the values in the MergedOptions."""
         for key in self.keys():
             yield self[key]
 
@@ -289,16 +398,20 @@ class MergedOptions(dict, Mapping):
         if converters is None:
             converters = self.converters
 
-        if isinstance(path, Path) and path.ignore_converters is ignore_converters and path.converters is converters:
+        path_type = type(path)
+        isPath = path_type is Path
+        if isPath and path.ignore_converters is ignore_converters and path.converters is converters:
             return path
 
-        joined = None
-        if hasattr(path, "joined"):
-            path, joined = path, path.joined()
-        elif isinstance(path, six.string_types):
-            path, joined = hp.prefixed_path_string(path, self.prefix_string)
-        elif isinstance(path, (list, tuple)):
-            path, joined = hp.prefixed_path_list(path, self.prefix_list)
+        if isPath:
+            joined = path.joined()
+        else:
+            from option_merge import helper as hp
+            if path_type in (list, tuple):
+                path, joined = hp.prefixed_path_list(path, self.prefix_list)
+            else:
+                path, joined = hp.prefixed_path_string(path, self.prefix_string)
+
         return Path.convert(path, self, converters=converters, joined=joined).ignoring_converters(ignore_converters)
 
     def add_converter(self, converter):
@@ -306,7 +419,30 @@ class MergedOptions(dict, Mapping):
         if converter not in self.converters:
             self.converters.append(converter)
 
+    def install_converters(self, converters, make_converter):
+        """
+        For each specified converter, make a converter function and install
+        a converter for that name.
+
+        .. code-block:: python
+
+            def make_converter(name, transformer):
+                def convert(path, val):
+                    return transformer(val)
+                return convert
+
+            m = MergedOptions.using({"a": 1, "b": 2})
+            m.install_converters({"a": lambda v: v+1, "b": lambda v: v*2}, make_converter)
+            m.converters.activate()
+
+            assert m['a'] == 2
+            assert m['b'] == 4
+        """
+        for name, spec in converters.items():
+            convert = make_converter(name, spec)
+            self.add_converter(Converter(convert=convert, convert_path=[name]))
+
     def as_dict(self, key="", ignore_converters=True, seen=None, ignore=None):
-        """Return this key as a single dictionary"""
+        """Collapse the storage at this prefix into a single dictionary"""
         return self.storage.as_dict(self.converted_path(key, ignore_converters=ignore_converters), seen=seen, ignore=ignore)
 
