@@ -8,20 +8,32 @@ which can be used to get keys and values for some path.
 It is also used to get thesource for particular paths.
 """
 
+from option_merge.versioning import versioned_iterable, versioned_value, VersionedDict
+from option_merge.merge import MergedOptions
+from option_merge.not_found import NotFound
+from option_merge.value_at import value_at
 from option_merge.joiner import dot_joiner
 from option_merge import helper as hp
 from option_merge.path import Path
 
-from delfick_error import ProgrammerError
-from collections import defaultdict
-from namedlist import namedlist
-
-class DataPath(namedlist("Path", ["path", "data", ("source", None)])):
+class DataPath(object):
     """
     Encapsulates a (path, data, source) triplet and getting keys and values from
     that path in the data.
     """
 
+    def __init__(self, path, data, source=None):
+        self.path = path
+        self.data = data
+        self.source = source
+
+    @property
+    def is_dict(self):
+        is_dict = getattr(self, "_is_dict", None)
+        if is_dict is None:
+            is_dict = self._is_dict = type(self.data) in (dict, VersionedDict, MergedOptions) or isinstance(self.data, dict)
+        return is_dict
+
     def items(self, prefix, want_one=False):
         """
         Yield (key, data, short_path) triplets for after this prefix
@@ -43,7 +55,10 @@ class DataPath(namedlist("Path", ["path", "data", ("source", None)])):
             raise NotFound
         """
         data = self.data
-        if not prefix or self.path.startswith("{0}.".format(prefix)) or self.path == prefix:
+        joined = self.path.joined()
+        joined_prefix = prefix.joined()
+
+        if not prefix or joined == joined_prefix or joined.startswith(joined_prefix + '.'):
             shortened_path = self.path.without(prefix)
             if shortened_path:
                 yield list(shortened_path)[0], data, list(shortened_path)[1:]
@@ -51,8 +66,8 @@ class DataPath(namedlist("Path", ["path", "data", ("source", None)])):
             else:
                 prefix = ""
         else:
-            if not dot_joiner(prefix).startswith(self.path.joined()):
-                raise hp.NotFound
+            if not joined_prefix.startswith(joined):
+                raise NotFound
             else:
                 prefix = Path(prefix).without(self.path)
 
@@ -61,9 +76,8 @@ class DataPath(namedlist("Path", ["path", "data", ("source", None)])):
                 yield "", data, []
                 return
 
-            from option_merge.merge import MergedOptions
-            if type(data) not in (dict, MergedOptions):
-                raise hp.NotFound
+            if type(data) not in (dict, VersionedDict, MergedOptions) and getattr(data, "is_dict", False) is not True:
+                raise NotFound
 
             if not prefix:
                 for key in data.keys():
@@ -79,13 +93,14 @@ class DataPath(namedlist("Path", ["path", "data", ("source", None)])):
                     break
 
             if not found:
-                raise hp.NotFound
+                raise NotFound
 
     def keys_after(self, prefix):
         """Yield the keys after this prefix"""
         for key, _, _ in self.items(prefix):
             yield key
 
+    @versioned_value
     def value_after(self, prefix):
         """Returns the value after prefix"""
         for key, data, short_path in self.items(prefix, want_one=True):
@@ -109,6 +124,7 @@ class Storage(object):
     def __init__(self):
         self.data = []
         self.deleted = []
+        self._version = -1
 
     ########################
     ###   USAGE
@@ -117,8 +133,9 @@ class Storage(object):
     def add(self, path, data, source=None):
         """Add data at the beginning"""
         if not isinstance(path, Path):
-            raise ProgrammerError("Path should be a Path object\tgot={0}".format(type(path)))
-        self.data.insert(0, (path, data, source))
+            raise Exception("Path should be a Path object\tgot={0}".format(type(path)))
+        self._version += 1
+        self.data.insert(0, (path, VersionedDict.convert(data), source))
 
     def get(self, path):
         """Get a single value from a path"""
@@ -136,37 +153,32 @@ class Storage(object):
         if not path:
             return []
 
-        for info in self.get_info(path, chain + [(path, self)], ignore_converters=True):
+        for info in self.get_info(path, ignore_converters=True, chain=chain + [(path, self)]):
             source = info.source
             if callable(info.source):
                 source = info.source()
 
-            if info.path == path and not isinstance(info.data, dict):
-                if isinstance(source, list):
-                    return list(source)
-                elif source:
-                    return [source]
-            else:
-                if source not in sources:
-                    if source:
-                        if isinstance(source, list):
-                            sources.extend(source)
-                        else:
-                            sources.append(source)
+            if info.path == path:
+                if source not in sources and source:
+                    if isinstance(source, list):
+                        for s in source:
+                            if s not in sources:
+                                sources.append(s)
+                    else:
+                        sources.append(source)
 
-        if len(sources) == 1:
-            return sources[0]
-        else:
-            return sources
+        return sources
 
     def delete(self, path):
         """Delete the first instance of some path"""
         for index, (info_path, data, _) in enumerate(self.data):
             dotted_info_path = dot_joiner(info_path)
             if dotted_info_path == path:
+                self._version += 1
                 del self.data[index]
                 return
             elif dotted_info_path.startswith("{0}.".format(path)):
+                self._version += 1
                 del self.data[index]
                 return
             elif not dotted_info_path or path.startswith("{0}.".format(dotted_info_path)):
@@ -182,7 +194,15 @@ class Storage(object):
     ###   IMPLEMENTATION
     ########################
 
-    def get_info(self, path, chain=None, ignore_converters=False):
+    @property
+    def version(self):
+        if self._version > 0:
+            return [self._version] + [getattr(item, "version", 0) for item in self.data]
+        else:
+            return -1
+
+    @versioned_iterable
+    def get_info(self, path, ignore_converters=False, chain=None):
         """Yield DataPath objects for this path in the data"""
         yielded = False
         if not self.data:
@@ -196,9 +216,8 @@ class Storage(object):
 
         for info_path, data, source in self.data:
             for full_path, found_path, val in self.determine_path_and_val(path, info_path, data, source):
-                source = self.make_source_for_function(val, found_path, chain, default=source)
-                path = Path.convert(path).ignoring_converters(ignore_converters)
-                yield DataPath(path.using(full_path), val, source)
+                source = self.make_source_for_function(data, found_path, chain, default=source)
+                yield DataPath(full_path, val, source)
                 yielded = True
 
         if not yielded:
@@ -210,39 +229,42 @@ class Storage(object):
 
         Where found_path is the path relative to the data
         """
-        path = Path.convert(path)
+        joined_path = dot_joiner(path)
+        joined_info_path = dot_joiner(info_path)
+        if joined_path == joined_info_path:
+            yield info_path, "", data
+            return
+
         try:
-            if path == info_path:
-                yield info_path, "", data
-                return
-
             if not info_path:
-                found_path, val = hp.value_at(data, path, self)
-                yield info_path + found_path, dot_joiner(found_path), val
+                found_path, val = value_at(data, Path.convert(path), self)
+                yield info_path + found_path, dot_joiner(found_path, list), val
                 return
 
-            if path.startswith("{0}.".format(dot_joiner(info_path))):
-                get_at = path.without(info_path)
-                found_path, val = hp.value_at(data, get_at, self)
+            if joined_path.startswith(joined_info_path + '.'):
+                get_at = Path.convert(path).without(info_path)
+                found_path, val = value_at(data, get_at, self)
                 yield info_path + found_path, dot_joiner(found_path), val
                 return
 
             # We are only part way into info_path
-            for key, data, short_path in DataPath(path.using(info_path), data, source).items(path, want_one=True):
+            path = Path.convert(path)
+            for key, data, short_path in DataPath(Path.convert(info_path), data, source).items(path, want_one=True):
                 yield path, "", hp.make_dict(key, short_path, data)
-        except hp.NotFound:
+        except NotFound:
             pass
 
-    def make_source_for_function(self, obj, path, chain, default=None):
+    def make_source_for_function(self, data, path, chain, default=None):
         """Return us a function that will get the source for some path on the specified obj"""
         def source_for():
-            if hasattr(obj, "source_for"):
-                nxt = obj.source_for(path, chain)
+            if hasattr(data, "source_for"):
+                nxt = data.source_for(data.converted_path(path), chain)
                 if nxt:
                     return nxt
             return default
         return source_for
 
+    @versioned_iterable
     def keys_after(self, path, ignore_converters=False):
         """Get all the keys after this path"""
         done = set()
@@ -252,25 +274,30 @@ class Storage(object):
                 continue
 
             try:
-                for key in info.keys_after(path):
-                    joined = dot_joiner([path, key])
+                for key in info.keys_after(Path.convert(path)):
+                    if path:
+                        joined = dot_joiner([path, key], list)
+                    else:
+                        joined = key
+
                     if not any(s == joined or joined.startswith("{0}.".format(s)) for s in stopped):
                         if key not in done:
                             yield key
                             done.add(key)
-            except hp.NotFound:
+            except NotFound:
                 pass
 
-            if not isinstance(info.data, dict):
+            if not info.is_dict:
                 stopped.add(dot_joiner(info.path))
 
     def delete_from_data(self, data, path):
         """Delete this path from the data"""
-        if not path or not isinstance(data, dict):
+        if not path or (type(data) not in (dict, VersionedDict, MergedOptions) and not isinstance(data, dict)):
             return False
 
         keys = list(reversed(sorted(data.keys())))
         if path in keys:
+            self._version += 1
             del data[path]
             return True
 
@@ -283,8 +310,10 @@ class Storage(object):
         """Return this path as a single dictionary"""
         result = {}
         if seen is None:
-            seen = defaultdict(list)
+            seen = {}
 
+        if path not in seen:
+            seen[path] = []
         if self in seen[path]:
             return {}
         seen[path].append(self)
@@ -293,44 +322,39 @@ class Storage(object):
             prefix, data, _ = self.data[i]
 
             if prefix:
-                prefix = list(prefix[:])
-                while prefix:
-                    data = {prefix.pop(): data}
-                if path:
-                    prefix, data = list(data.items())[0]
-                else:
-                    prefix = list(data.keys())[0]
-
-            try:
-                if path:
-                    path_without_prefix = path.without(prefix)
-                else:
-                    path_without_prefix = path
-            except hp.NotFound:
-                continue
+                prefixer = list(prefix)
+                while prefixer:
+                    key = prefixer.pop()
+                    data = VersionedDict({key: data})
 
             val = data
             used = None
             found = False
+            path_without_prefix = path
             while not found or path_without_prefix:
-                if hasattr(data, "as_dict"):
-                    val = val.as_dict(path_without_prefix, seen=seen, ignore=ignore)
-                    path_without_prefix = Path([])
+                if hasattr(data, "is_dict") and data.is_dict:
+                    if hasattr(val, "as_dict"):
+                        val = val.as_dict(path_without_prefix, seen=seen, ignore=ignore)
+                        path_without_prefix = ""
+                    else:
+                        val = dict(val)
 
                 try:
-                    used, val = hp.value_at(val, path_without_prefix, self)
+                    used, val = value_at(val, path_without_prefix, self)
                     found = True
-                except hp.NotFound:
+                except NotFound:
                     found = False
                     break
 
-                path_without_prefix = path_without_prefix.without(dot_joiner(used))
+                if used and path_without_prefix:
+                    path_without_prefix = path_without_prefix.without(dot_joiner(used))
 
             if found and path_without_prefix == "":
-                if not isinstance(val, dict):
+                is_dict = lambda item: type(item) in (dict, VersionedDict, MergedOptions) or isinstance(item, dict)
+                if not is_dict(val):
                     result = val
                 else:
-                    if not isinstance(result, dict):
+                    if not is_dict(result):
                         result = {}
                     hp.merge_into_dict(result, val, seen, ignore=ignore)
 
